import re
from typing import List

from lib.paraver import *
from lib.ParaverLine import *


class DisasmRV:
    last_inst_without_rdest = True
    last_inst_with_addr = False
    last_inst_setvec = False
    last_inst_masked = False
    last_inst_overwrite_src = False

    current_inst_setvec = False
    current_inst_without_rdest = False
    current_rdest = 0
    current_inst_with_addr = False
    current_inst_masked = False
    current_inst_overwrite_src = False
    current_inst_tuple = False
    current_impl_rdest = 0


    insts_without_rdest = ["vse8", "vse16", "vse32", "vse64", "vs1r"]
    insts_with_addr = ["vse8", "vse16", "vse32", "vse64", "vle8", "vle16", "vle32", "vle64", "vs1r", "vl1r"]
    insts_setvec = ["vsetli"]
    insts_tuple = ["vzip2", "vunzip2", "vtrn"]


    #TODO: erase magic numbers
    imm_index = [0, 1]
    vs_index = [2, 3, 4]
    inst_index = [5]
    raddr_index = [6]
    immaddr_index = [7]
    mask_index = [8]
    inst_overwrite_index = [9]
    vlen_index = [13]
    elen_index = [14]


    re_immediate = re.compile("(\s|,)\s*(0x)?([0-9]+)\s*(,|\(|$)")
    re_addr = re.compile("((0x)?[0-9]*)?\((.*)\)")
    re_inst = re.compile("^(\w*)(\.|\t|\n|\s|$)")
    re_mask = re.compile("(v[0-9]+)\.t")
    re_vreg = re.compile("v([0-9]+)")

    def __init__(self):
        pass

    def userEvents(self, row, paraverEvents, paraver_line):
        if row.startswith(".trace"):
            row = row[len(".trace"):];

            row_split = row.split(",")
            usr_event = PrvEvent("", int(row_split[0]))

            paraver_line.addEvent(paraverEvents[self.inst_index[0]], 0)
            paraver_line.addEvent(usr_event, int(row_split[1]))
        elif row.startswith(".config "):
            row = row[len(".config "):];
            # Parse VLEN and ELEN
            row_split = row.split(",")
            assert len(row_split) == 2
            row_split = list(map(lambda x : x.split("="), row_split))
            assert row_split[0][0] == "VLEN"
            assert row_split[1][0] == "ELEN"
            paraver_line.addEvent(paraverEvents[self.vlen_index[0]], int(row_split[0][1]))
            paraver_line.addEvent(paraverEvents[self.elen_index[0]], int(row_split[1][1]))
        else:
            assert False

    def getInstructionArguments(self, disasm):

        arguments = disasm.split("\t")[1:]

        arguments = arguments[0].split(",")
        # arguments = [x for x in arguments if not x=='' ]
        arguments = [x[:-2] if x[-1] == ',' else x for x in arguments]

        return arguments

    def getValuesHashedFromDisasm(self, disasm, paraverEvents, paraverLine, start, end, add_new=False):
        position = start
        args = self.getInstructionArguments(disasm)

        for a in args:
            a = a.replace(" ", "")
            a = a.replace("(", "")
            a = a.replace(")", "")

            if not a in paraverEvents[position].values:
                if not add_new:
                    continue

                num_values = len(paraverEvents[position].values)
                paraverEvents[position].addValue(num_values + 1, a)

            paraverLine.addEvent(paraverEvents[position], paraverEvents[position].values[a])
            position = position + 1

        return position


    def getImmediateFromDisasm(self, disasm, paraverEvents, paraverLine):
        match = self.re_immediate.search(disasm)
        if match is not None:
            imm = match.group(3)
            paraverLine.addEvent(paraverEvents[self.imm_index[0]], imm)

    def getAddrFromDisasm(self, disasm, paraverevents: PrvEvent, paraverline: ParaverLine, raddr_index, imm_index):
        match = self.re_addr.search(disasm)
        if match is not None:
            imm = match.groups()[0]
            reg = match.groups()[2]
            if imm == "":
                imm = 0
            else:
                imm = int(imm, 0)
            paraverline.addEvent(paraverevents[self.raddr_index[0]], paraverevents[raddr_index].values[reg])
            paraverline.addEvent(paraverevents[self.imm_index[0]], imm)

    def getInstruction(self, disasm):
        return self.re_inst.search(disasm).group(1)

    def getMaskFromDisasm(self, disasm, paraver_event: PrvEvent, paraver_line: ParaverLine):

        match = self.re_mask.search(disasm)

        if match is not None:
            mask_reg = match.group(1)
            value = paraver_event.values[mask_reg]
            paraver_line.addEvent(paraver_event, value)
            return True
        else:
            return False

    def inference_null_values_setvec(self, paraverEvents, paraverLine):
        if not self.current_inst_setvec and self.last_inst_setvec:
            for i in self.vs_index:
                paraverLine.addEvent(paraverEvents[i], 0)
        self.last_inst_setvec = self.current_inst_setvec

    def inference_null_values_with_address(self, paraverEvents, paraverLine):
        if not self.current_inst_with_addr and self.last_inst_with_addr:
            for i in self.raddr_index + self.immaddr_index:
                paraverLine.addEvent(paraverEvents[i], 0)
        self.last_inst_with_addr = self.current_inst_with_addr

    def inference_null_values_with_mask(self, paraverEvents, paraverLine):
        if not self.current_inst_masked and self.last_inst_masked:
            paraverLine.addEvent(paraverEvents[self.mask_index[0]], 0)
        self.last_inst_masked = self.current_inst_masked

    def inference_null_values_overwrite(self, paraver_events: List[PrvEvent], paraver_line: ParaverLine):
        if not self.current_inst_overwrite_src and self.last_inst_overwrite_src:
            paraver_line.addEvent(paraver_events[self.inst_overwrite_index[0]], 0)
        self.last_inst_overwrite_src = self.current_inst_overwrite_src

    def inference_null_values(self, paraver_events: List[PrvEvent], paraver_line: ParaverLine):
        self.inference_null_values_setvec(paraver_events, paraver_line)
        self.inference_null_values_with_address(paraver_events, paraver_line)
        self.inference_null_values_with_mask(paraver_events, paraver_line)
        self.inference_null_values_overwrite(paraver_events, paraver_line)

    def disasmParser(self, disasm, paraverEvents, paraverLine):
        inst = self.getInstruction(disasm)
        paraverLine.addEvent(paraverEvents[self.inst_index[0]], paraverEvents[self.inst_index[0]].values[inst])

        # Parse setvec instructions
        if inst in self.insts_setvec:
            self.getValuesHashedFromDisasm(disasm, paraverEvents, paraverLine, self.vs_index[0], self.vs_index[-1],
                                           False)
            self.current_inst_setvec = True
        else:
            self.current_inst_setvec = False

        # Parse instructions with addresses
        if inst in self.insts_with_addr:
             self.getAddrFromDisasm(disasm, paraverEvents, paraverLine, self.raddr_index[0], self.immaddr_index[0])
             self.current_inst_with_addr = True
        else:
            self.current_inst_with_addr = False

        self.current_inst_masked = self.getMaskFromDisasm(disasm, paraverEvents[self.mask_index[0]], paraverLine)
        self.getImmediateFromDisasm(disasm, paraverEvents, paraverLine)

    def neutral_state(self):
        self.current_rdest = 0
        self.current_inst_without_rdest = True
        self.current_inst_with_addr = False
        self.current_inst_setvec = False
        self.current_inst_masked = False
        self.current_inst_overwrite_src = False
        self.current_inst_tuple = False

    def disasmToPRV(self, disasm, paraver_events, paraver_line):
        if disasm[0] == '.':
            self.userEvents(disasm, paraver_events, paraver_line)
            self.neutral_state()
        else:
            self.disasmParser(disasm, paraver_events, paraver_line)

        self.inference_null_values(paraver_events, paraver_line)
